"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

import click

from hcs_core.ctxp import CtxpException, recent
from hcs_core.ctxp.cli_options import apply_env as apply_env
from hcs_core.ctxp.cli_options import confirm as confirm
from hcs_core.ctxp.cli_options import env as env
from hcs_core.ctxp.cli_options import exclude_field as exclude_field
from hcs_core.ctxp.cli_options import field as field
from hcs_core.ctxp.cli_options import first as first
from hcs_core.ctxp.cli_options import force as force
from hcs_core.ctxp.cli_options import formatter as formatter
from hcs_core.ctxp.cli_options import ids as ids
from hcs_core.ctxp.cli_options import limit as limit
from hcs_core.ctxp.cli_options import output as output
from hcs_core.ctxp.cli_options import search as search
from hcs_core.ctxp.cli_options import sort as sort
from hcs_core.ctxp.cli_options import verbose as verbose
from hcs_core.ctxp.cli_options import wait as wait

org_id = click.option(
    "--org",
    type=str,
    default=None,
    required=False,
    help="Specify org ID. If not specified, default to the currently in-use org.",
)


def get_org_id(org: str = None) -> str:
    # 1st priority: user explicitly specified org
    with recent.of("org") as r:
        try:
            if org:
                return org

            # 2nd priority: environment variable
            org = os.environ.get("HCS_ORG")
            if org:
                return org

            # 3rd priority: last used
            org = r.get()
            if org:
                return org

            # 4th priority: org from the current auth token
            from hcs_core.sglib import auth

            auth_info = auth.details(False)
            if not auth_info:
                raise CtxpException("Not authorized. See 'hcs login --help'.")
            org = auth_info["org"]["id"]
            return org
        finally:
            if org:
                r.set(org)


def ensure_login():
    return get_org_id(None)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import time
from importlib.metadata import version

import httpx
from packaging.version import Version

import hcs_core

log = logging.getLogger(__name__)


def check_upgrade():
    last_upgrade_check_at = "last_upgrade_check_at"
    checked_at = hcs_core.ctxp.state.get(last_upgrade_check_at, 0)

    now = time.time()
    if now - checked_at < 24 * 60 * 60:
        return

    try:
        latest = get_latest_version()
        current = Version(get_version())
        if current < latest:
            log.warning(f"New version available: {latest}. Execute 'hcs upgrade' to upgrade.")
    except Exception as e:
        logging.debug(e)

    hcs_core.ctxp.state.set(last_upgrade_check_at, now)


def get_version():
    return version("hcs-cli")


def get_latest_version() -> Version:
    res = httpx.get("https://pypi.org/pypi/hcs-cli/json")
    names = res.json().get("releases").keys()
    versions = [Version(n) for n in names]
    versions.sort(reverse=True)
    return versions[0]


def is_version(s: str):
    try:
        return Version(s)
    except ValueError:
        return False



import json
import logging
import sys
import time
from datetime import datetime, timezone

import click
import httpx
from yumako import env

log = logging.getLogger(__name__)

_record = None
_enabled = None
_app_name = ""


def disable():
    global _enabled
    _enabled = False


def _is_disabled():
    global _enabled
    if _enabled is None:
        _enabled = env.bool("HCS_CLI_TELEMETRY", True)
    return not _enabled


def _get_version():
    try:
        from importlib.metadata import version

        return version("hcs-cli")
    except Exception as e:
        log.debug(f"Failed to get hcs-cli version: {e}")
        return "unknown"


def _get_record():
    global _record
    if _record is None:
        _record = {
            "@timestamp": datetime.now(timezone.utc).isoformat(timespec="milliseconds"),
            "app": _app_name,
            "command": None,
            "options": [],
            "return": -1,
            "error": None,
            "time_ms": -1,
            "version": _get_version(),
            "env": {
                "python_version": sys.version,
                "platform": sys.platform,
                "executable": sys.executable,
            },
        }
    return _record


def start(app_name: str = None):
    if _is_disabled():
        return

    global _app_name
    _app_name = app_name
    _get_record()


def update(cmd_path: str, params: dict):
    if _is_disabled():
        return

    record = _get_record()
    record["command"] = cmd_path
    record["options"] = [k.replace("_", "-") for k, v in params.items() if v]


def end(return_code: int = 0, error: Exception = None):
    if _is_disabled():
        return

    record = _get_record()
    if error:
        if isinstance(error, click.exceptions.Exit):
            return_code = error.exit_code
        elif isinstance(error, SystemExit):
            return_code = error.code
        else:
            record["error"] = str(error)
            if return_code == 0:
                return_code = 1
    record["return"] = return_code
    record["time_ms"] = int((time.time() - datetime.fromisoformat(record["@timestamp"]).timestamp()) * 1000)

    _fix_missing_commands(record)
    _injest(record)
    return record


def _fix_missing_commands(record):
    if record["command"]:
        return

    args = sys.argv[1:]

    # this does not work for all cases, but only as best effort.
    options_started = False
    options = record["options"]
    command = [_app_name]
    for arg in args:
        if arg.startswith("-"):
            options_started = True

        if options_started:
            if arg.startswith("--"):
                options.append(arg[2:])
            elif arg.startswith("-"):
                options.append(arg[1:])
            else:
                # value. For privacy no logging.
                continue
        else:
            command.append(arg)

    record["command"] = " ".join(command)


def _injest(doc):
    # print('TELEMETRY end', json.dumps(doc, indent=4), flush=True)

    try:
        response = httpx.post(
            "https://collie.omnissa.com/es/hcs-cli/_doc",
            auth=("append_user", "public"),
            headers={"Content-Type": "application/json"},
            content=json.dumps(doc),
            timeout=4,
            verify=False,
        )
        response.raise_for_status()
    except Exception as e:
        log.debug(f"Telemetry ingestion failed: {e}", exc_info=True)
        return



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# nanw @ 2023
# PKCE with authlib: https://docs.authlib.org/en/latest/specs/rfc7636.html

import logging
import platform
import secrets
import threading
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import parse_qs, urlparse

from authlib.integrations.httpx_client import OAuth2Client
from authlib.oauth2.rfc7636 import create_s256_code_challenge

log = logging.getLogger(__name__)

_server_address = ("127.0.0.1", 10762)
_callback_url = "/hcs-cli/oauth/callback"

_public_client_ids = {
    "production": "ldjmWbBAUcSB1w3HSbzoYcdKoloYFqT2dWK",
    # "production": "WbvEYXK4abljPJUuY2zODLpOaX5ddH2uhMX",    #collie mobile
    "staging": "BbwHlgWt0vFwPUZMJTb5IKltynXjDmb46IO",
}

_auth_code_event = threading.Event()
_auth_code_event.value = None
_state = None

_auth_success_html = """
<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8" />
    <title>Login successfully</title>
    <style>
        body {
            font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
        }

        code {
            font-family: Consolas, 'Liberation Mono', Menlo, Courier, monospace;
            display: inline-block;
            background-color: rgb(242, 242, 242);
            padding: 12px 16px;
            margin: 8px 0px;
        }
    </style>
</head>
<body>
    <h3>You have successfully logged into Omnissa Horizon Cloud Service.</h3>
    <p>You can close this window, and return to the terminal.</p>
    <br/>
    <p><a href="https://github.com/euc-eng/hcs-cli/blob/dev/README.md">HCS CLI</a> is in tech preview. <a href="https://github.com/euc-eng/hcs-cli/blob/main/doc/hcs-cli-cheatsheet.md">Cheatsheet</a>:</p>
    <code>
    # To get the login details: <br/>
    hcs login -d <br/><br/>
    # To login programmatically: <br/>
    hcs login --api-token &lt;my-CSP-user-API-token&gt; <br/><br/>
    # Switch to a different profile:<br/>
    hcs profile use <br/><br/>
    </code>
</body>
</html>
"""


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        def _respond(code: int, reason: str, content_type: str = "text/plain"):
            self.send_response(code)
            self.send_header("Content-type", content_type)
            self.end_headers()
            self.wfile.write(reason.encode())
            self.wfile.flush()

        try:
            # Parse the query parameters from the callback URL
            parsed_url = urlparse(self.path)
            log.debug(parsed_url)
            if parsed_url.path == _callback_url or parsed_url.path == "/hcsadmin/index.html":
                query_components = parse_qs(parsed_url.query)
                codes = query_components.get("code", None)
                states = query_components.get("state", None)

                # Check if the 'code' parameter is present in the callback URL
                if not codes:
                    _respond(400, "Missing authorization code")
                    return
                if not states:
                    _respond(400, "Missing state in server response")
                    return
                state = states[0]
                if _state != state:
                    _respond(401, "Response state mismatch")
                    return

                code = codes[0]
                # Process the authorization code
                log.debug(f"auth code {code}")
                _auth_code_event.value = code

                # Send a response back to the client
                _respond(200, _auth_success_html, "text/html")
            else:
                _respond(404, "Not found")
        finally:
            # Terminate the server
            _auth_code_event.set()
            threading.Thread(target=lambda: self.server.shutdown()).start()

    def log_message(self, *args):
        log.debug(*args)


def run_server_async():
    def run_server():
        log.debug("oauth callback httpd - start")
        httpd = HTTPServer(_server_address, OAuthCallbackHandler)
        httpd.serve_forever()
        log.debug("oauth callback httpd - exit")

    threading.Thread(target=run_server, daemon=True).start()


def do_oauth2_pkce(csp_url: str, client_id: str, org_id: str):
    # OAuth2 server details
    # authorization_endpoint = csp_url + '/csp/gateway/am/api/auth/api-tokens/authorize'
    discovery_endpoint = csp_url + "/csp/gateway/discovery"
    token_endpoint = csp_url + "/csp/gateway/am/api/auth/token"

    redirect_uri = f"http://{_server_address[0]}:{_server_address[1]}{_callback_url}"

    # Create an OAuth2Session instance
    session = OAuth2Client(client_id, redirect_uri=redirect_uri)

    # Generate a code verifier and code challenge

    code_verifier = secrets.token_urlsafe(64)
    code_challenge = create_s256_code_challenge(code_verifier)
    # code_challenge = CodeChallenge().from_code_verifier(code_verifier)

    # Create authorization URL
    # https://console-stg.cloud.vmware.com/csp/gateway/authn/api/swagger-ui.html#/Discovery/getDiscoveryUsingGET
    global _state
    authorization_url, _state = session.create_authorization_url(
        discovery_endpoint, orgId=org_id, code_challenge=code_challenge, code_challenge_method="S256"
    )

    log.debug(authorization_url)

    # Redirect the user to the authorization URL
    webbrowser.open(authorization_url, new=0, autoraise=True)

    try:
        # On Windows, wait() without timeout doesn't respond to CTRL+C properly
        # Use a polling loop with timeout to allow KeyboardInterrupt to be caught
        if platform.system() == "Windows":
            while not _auth_code_event.wait(timeout=0.5):
                pass  # Keep waiting until the event is set
        else:
            _auth_code_event.wait()
    except KeyboardInterrupt:
        log.info("Aborted by user")
        return
    except Exception as e:
        log.error(f"Login error: {e}")
        return
    # Once the user is redirected back to your app with an authorization code, exchange it for an access token
    authorization_code = _auth_code_event.value
    log.debug("authorization_code: %s", authorization_code)
    if not authorization_code:
        log.debug("Login failed")
        return

    token = session.fetch_token(
        token_endpoint,
        auth=(client_id, ""),  # auth method required by CSP API for PKCE flow
        authorization_response=redirect_uri + "?code=" + authorization_code,
        code_verifier=code_verifier,
    )

    log.debug("token: %s", token)
    return token


def identify_client_id(csp_url: str) -> str:
    if csp_url.find("console.cloud") >= 0:
        return _public_client_ids["production"]
    if csp_url.find("https://connect.omnissa.com") >= 0:
        return _public_client_ids["production"]

    if csp_url.find("console-stg.cloud") >= 0:
        return _public_client_ids["staging"]
    raise Exception("Unknonw CSP url: " + csp_url)


def login_via_browser(csp_url: str, org_id: str):
    client_id = identify_client_id(csp_url)
    run_server_async()
    return do_oauth2_pkce(csp_url, client_id, org_id)


# def _test():
#     import jwt
#     import json
#     csp_url = 'https://console-stg.cloud.vmware.com'
#     org_id = '06e09ab5-dc8d-413c-bd57-2c1e929a412c' #stg horizonv2-syntest
#     token = interactive_login_via_browser(csp_url, org_id)
#     decoded = jwt.decode(token['access_token'], options={"verify_signature": False})
#     print(json.dumps(decoded, indent=4))

# if __name__ == '__main__':
#     _test()



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
import sys
import threading
import time
from typing import Callable, Iterator

from hcs_core.ctxp import CtxpException, panic, profile
from hcs_core.sglib.ez_client import EzClient
from hcs_core.util import duration, exit
from hcs_core.util.query_util import PageRequest, with_query

log = logging.getLogger(__name__)

_caches = {}

_client_instance_lock = threading.RLock()


def _get_service_override(service_name: str):
    profile_data = profile.current()
    override = profile_data.get("override", {})
    service_override = {}
    for k, v in override.items():
        if service_name == k.lower():
            service_override = v

    if not service_override:
        if service_name == "org-service":
            service_override = override.get("org", {})
        elif service_name.find("-") >= 0:
            camel_name = "".join(word.capitalize() for word in service_name.split("-"))
            camel_name = camel_name[0].lower() + camel_name[1:]
            service_override = override.get(camel_name, {})
    return service_override


def _lazy_init(service_name: str, hdc: str = None, region: str = None):  # make it deferred so no need to initialize profile
    if region and hdc:
        raise Exception("region and hdc cannot be specified at the same time.")

    profile_data = profile.current()

    service_override = _get_service_override(service_name)
    service_override_url = service_override.get("url")
    if service_override_url:
        log.debug(f"Using per-service override for {service_name}: {service_override_url}")
        url = service_override_url
    else:
        if hdc:
            # prod only
            hdc = hdc.upper()
            if hdc == "US":
                url = profile_data.hcs.url
            elif hdc == "EU":
                url = profile_data.hcs.url.replace("cloud-sg-us", "cloud-sg-eu")
            elif hdc == "JP":
                url = profile_data.hcs.url.replace("cloud-sg-us", "cloud-sg-jp")
            else:
                raise CtxpException(f"Invalid HDC name: {hdc}. Supported HDC names: US, EU, JP.")
        elif region:
            url = _get_region_url(region)
            if not url:
                panic("Missing profile property: hcs.regions")
        else:
            url = profile_data.hcs.url
    url = url.rstrip("/") + "/" + service_name

    # TODO
    client_id = service_override.get("client-id", None)
    if not client_id:
        client_id = service_override.get("clientId", None)
    client_secret = service_override.get("client-secret", None)
    if not client_secret:
        client_secret = service_override.get("clientSecret", None)
    api_token = service_override.get("api-token", None)
    if not api_token:
        api_token = service_override.get("apiToken", None)
    provider = service_override.get("provider", "vmwarecsp")
    if provider == "vmwarecsp":
        token_url = profile_data.csp.url
    elif provider == "auth-service":
        token_url = profile_data.auth["tokenUrl"]
    else:
        raise CtxpException(f"Unknown provider: {provider}. Supported providers: vmwarecsp, auth-service.")

    if client_id:
        if not client_secret:
            panic(f"Client ID is specified but missing clientSecret for service {service_name} in profile override.")
    else:
        if client_secret:
            panic(f"Client secret is specified but missing clientId for service {service_name} in profile override.")

    if client_id:
        auth_config = {
            "url": token_url,
            "client_id": client_id,
            "client_secret": client_secret,
        }
    elif api_token:
        auth_config = {
            "url": token_url,
            "api_token": api_token,
        }
    else:
        auth_config = None

    from hcs_core.sglib import auth

    oauth_client = auth.oauth_client(auth_config=auth_config)
    return url, oauth_client


def hdc_service_client(service_name: str, hdc: str = None) -> EzClient:
    """A client for HDC service. Cached per service name, thread-safe, lazy initialization."""
    if hdc:
        hdc = hdc.upper()
        if hdc not in ["US", "EU", "JP"]:
            panic(f"Invalid HDC name: {hdc}. Supported HDC names: US, EU, JP.")
    else:
        hdc = "US"

    k = f"{service_name}#{hdc}"
    with _client_instance_lock:
        instance = _caches.get(k)
        if not instance:
            instance = EzClient(lazy_init=lambda: _lazy_init(service_name=service_name, hdc=hdc))
            _caches[k] = instance
        return instance


def _get_region_url(region: str):
    regions = profile.current().hcs.regions
    if not region:
        return regions[0].url
    for r in regions:
        if r.name.lower() == region.lower():
            return r.url
    names = []
    for r in regions:
        names.append(r.name)
    panic(f"Region not found: {region}. Available regions: {names}")


def regional_service_client(service_name: str, region: str = None):
    # 'https://dev1b-westus2-cp103a.azcp.horizon.vmware.com/vmhub'
    region_names = [r.name.lower() for r in profile.current().hcs.regions]
    if region:
        region = region.lower()
        if region not in region_names:
            panic(f"Invalid region name: {region}. Available regions: {', '.join(region_names)}")
    else:
        region = region_names[0]

    k = f"{service_name}#{region}"
    with _client_instance_lock:
        instance = _caches.get(k)
        if not instance:
            instance = EzClient(lazy_init=lambda: _lazy_init(service_name=service_name, region=region))
            _caches[k] = instance
        return instance


def is_regional_service(service_name: str):
    regional_services = ["vmhub", "connection-service"]
    return service_name in regional_services


def service_client(service_name: str, region: str = None, hdc: str = None):
    if is_regional_service(service_name):
        return regional_service_client(service_name, region)
    else:
        return hdc_service_client(service_name, hdc)


class default_crud:
    def __init__(self, client, base_context: str, resource_type_name: str):
        self._client_impl = client
        self._base_context = base_context
        self._resource_type_name = resource_type_name

    def _client(self):
        if callable(self._client_impl):
            self._client_impl = self._client_impl()
        elif isinstance(self._client_impl, str):
            self._client_impl = hdc_service_client(self._client_impl)
        else:
            pass
        if isinstance(self._client_impl, EzClient):
            return self._client_impl
        raise CtxpException(f"Invalid client implementation: {self._client_impl}")

    def get(self, id: str, org_id: str, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id
        url = with_query(f"{self._base_context}/{id}", **kwargs)
        # print(url)
        return self._client().get(url)

    def list(self, org_id: str, fn_filter: Callable = None, **kwargs) -> list:
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id

        def _get_page(query_string):
            url = self._base_context + "?" + query_string
            # print(url)
            return self._client().get(url)

        return PageRequest(_get_page, fn_filter, **kwargs).get()

    def items(self, org_id: str, fn_filter: Callable = None, **kwargs) -> Iterator:
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id

        def _get_page(query_string):
            url = self._base_context + "?" + query_string
            return self._client().get(url)

        return PageRequest(_get_page, fn_filter, **kwargs).items()

    def create(self, payload: dict, headers: dict = None, **kwargs):
        url = with_query(f"{self._base_context}", **kwargs)
        # print(url)
        # import json
        # print(json.dumps(payload, indent=4))
        if isinstance(payload, str):
            return self._client().post(url, text=payload, headers=headers)
        if isinstance(payload, dict):
            return self._client().post(url, json=payload, headers=headers)
        return self._client().post(url, json=payload, headers=headers)

    def upload(self, files, **kwargs):
        url = with_query(f"{self._base_context}", **kwargs)
        return self._client().post(url, files=files)

    def delete(self, id: str, org_id: str, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id
        url = with_query(f"{self._base_context}/{id}", **kwargs)
        # print(url)
        return self._client().delete(url)

    def wait_for_deleted(self, id: str, org_id: str, timeout: str, fn_is_error: Callable = None):
        name = self._resource_type_name + "/" + id

        def fn_get():
            return self.get(id, org_id)

        return wait_for_res_deleted(name, fn_get, timeout=timeout, fn_is_error=fn_is_error)

    def update(self, id: str, org_id: str, data: dict, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
            kwargs["orgId"] = org_id
        url = with_query(f"{self._base_context}/{id}")
        return self._client().patch(url, data)


def _parse_timeout(timeout: str):
    if isinstance(timeout, int):
        return timeout
    if isinstance(timeout, str):
        return duration.to_seconds(timeout)

    raise CtxpException(f"Invalid timout. Type={type(timeout).__name__}, value={timeout}")


def wait_for_res_deleted(
    resource_name: str,
    fn_get: Callable,
    timeout: str,
    polling_interval: int = 10,
    fn_is_error: Callable = None,
):
    timeout_seconds = _parse_timeout(timeout)
    polling_interval_seconds = _parse_timeout(polling_interval)
    if polling_interval_seconds < 3:
        polling_interval_seconds = 3
    start = time.time()
    while True:
        t = fn_get()
        if t is None:
            return
        if fn_is_error:
            if fn_is_error(t):
                msg = f"Failed deleting resource '{resource_name}', resource in Error state."
                raise CtxpException(msg)

        now = time.time()
        remaining_seconds = timeout_seconds - (now - start)
        if remaining_seconds < 1:
            msg = f"Timeout waiting for resource '{resource_name}' to be deleted."
            raise TimeoutError(msg)
        sleep_seconds = remaining_seconds
        if sleep_seconds > polling_interval_seconds:
            sleep_seconds = polling_interval_seconds
        time.sleep(sleep_seconds)


# flake8: noqa: E731
def wait_for_res_status(
    resource_name: str,
    fn_get: Callable,
    get_status: Callable,
    status_map: dict = None,
    is_ready: Callable = None,
    is_error: Callable = None,
    is_transition: Callable = None,
    timeout: str = "10m",
    polling_interval: str = "20s",
    not_found_as_success: bool = False,
):
    timeout_seconds = _parse_timeout(timeout)
    polling_interval_seconds = _parse_timeout(polling_interval)
    if polling_interval_seconds < 3:
        polling_interval_seconds = 3
    start = time.time()
    prefix = f"Error waiting for resource {resource_name}: "

    if isinstance(get_status, str):
        field_name = get_status
        get_status = lambda t: t[field_name]
    if status_map:
        if is_ready:
            raise CtxpException("Can not specify is_ready when status_map is provided.")
        if is_error:
            raise CtxpException("Can not specify is_error when status_map is provided.")
        if is_transition:
            raise CtxpException("Can not specify is_transition when status_map is provided.")

        ready_status = status_map["ready"]
        error_status = status_map["error"]
        transition_status = status_map["transition"]
        if isinstance(ready_status, str):
            ready_status = [ready_status]
        if isinstance(error_status, str):
            error_status = [error_status]
        if isinstance(transition_status, str):
            transition_status = [transition_status]
        is_ready = lambda s: s in ready_status
        is_error = lambda s: error_status is None or s in error_status
        is_transition = lambda s: transition_status is None or s in transition_status
    else:
        if not is_ready:
            raise CtxpException("Either status_map or is_ready must be specified.")
        if not is_error:
            raise CtxpException("Either status_map or is_error must be specified.")
        if not is_transition:
            raise CtxpException("Either status_map or is_transition must be specified.")
        ready_status = None
        error_status = None
        transition_status = None

    while True:
        t = fn_get()
        if t is None:
            if not_found_as_success:
                return
            raise CtxpException(prefix + "Not found.")
        status = get_status(t)
        if is_error(status):
            msg = prefix + f"Unexpected terminal state. Actual={status}"
            if ready_status:
                msg += f", expected={ready_status}"
            print("-- DUMP START --", file=sys.stderr)
            print(json.dumps(t, indent=4), file=sys.stderr)
            print("-- DUMP END --", file=sys.stderr)
            raise CtxpException(msg)
        if is_ready(status):
            return t
        if not is_transition(status):
            raise CtxpException(prefix + f"Unexpected status: {status}. If this is a transition, add it to status_map['transition'].")

        now = time.time()
        remaining_seconds = timeout_seconds - (now - start)
        if remaining_seconds < 1:
            msg = prefix + f"Timeout. Current: {status}."
            if ready_status:
                msg += f" Expected: {ready_status}."
            raise TimeoutError(msg)
        sleep_seconds = remaining_seconds
        if sleep_seconds > polling_interval_seconds:
            sleep_seconds = polling_interval_seconds

        exit.sleep(sleep_seconds)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

import click

verbose = click.option(
    "--verbose",
    "-v",
    count=True,
    default=0,
    hidden=True,
    help="Print debug logs",
)

output = click.option(
    "--output",
    "-o",
    type=click.Choice(["json", "json-compact", "yaml", "yml", "table", "t", "text"], case_sensitive=False),
    default=None,
    hidden=True,
    help="Specify output format",
)

field = click.option(
    "--field",
    type=str,
    required=False,
    hidden=True,
    help="Specify fields to output, in comma separated field names.",
)

exclude_field = click.option(
    "--exclude-field",
    type=str,
    required=False,
    hidden=True,
    help="Specify fields to exclude from output, in comma separated field names.",
)

wait = click.option(
    "--wait",
    "-w",
    type=str,
    is_flag=False,
    flag_value="1m",
    default="0",
    help="Wait time. E.g. '30s', or '5m'. Default: 1m.",
)

search = click.option(
    "--search",
    "-s",
    type=str,
    required=False,
    help="Specify REST search. E.g. 'name $eq something'. Note: use single quote in bash/sh.",
)

sort = click.option(
    "--sort",
    type=str,
    required=False,
    help="Ascending/Descending. Format is property,{asc|desc} and default is ascending",
)


def limit(func=None, *, default=100):
    decorator = click.option(
        "--limit", "-l", type=int, required=False, default=default, help="Optionally, specify the number of records to fetch."
    )
    if func is not None:
        # Called as @limit (without parentheses)
        return decorator(func)
    else:
        # Called as @limit() or @limit(default=...)
        return decorator


ids = click.option(
    "--ids",
    "--id-only",  # to be removed.
    is_flag=True,
    type=bool,
    required=False,
    default=False,
    hidden=True,
    help="Return only id of the document",
)

first = click.option(
    "--first",
    is_flag=True,
    type=bool,
    required=False,
    default=False,
    hidden=True,
    help="Return only the first object, if the result is a list.",
)

force = click.option("--force/--grace", type=bool, default=True, help="Specify deletion mode: forceful, or graceful.")

confirm = click.option("--confirm/--prompt", "-y", type=bool, default=False, help="Confirm the operation without prompt.")

env = click.option(
    "--env",
    multiple=True,
    type=str,
    required=False,
    help="Alternative explicit in-line environment variable override in KEY=VALUE format.",
)


def apply_env(envs):
    for kv in envs:
        k, v = kv.split("=", 1)
        os.environ[k.strip()] = v.strip()


def formatter(formatter=None, columns=None):
    def decorator(f):
        f.formatter = formatter if formatter else columns
        return f

    return decorator



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import hashlib
import json
import logging
import threading
import time

import jwt
from authlib.integrations.httpx_client import OAuth2Client

from hcs_core.ctxp import CtxpException, profile
from hcs_core.ctxp.jsondot import dotdict, dotify

from .csp import CspClient

log = logging.getLogger(__name__)


def _is_auth_valid(auth_data):
    leeway = 60
    return auth_data and time.time() + leeway < auth_data["expires_at"]


_login_lock = threading.Lock()


def login(force_refresh: bool = False, verbose: bool = False):
    """Ensure login state, using credentials from the current profile. Return oauth token."""
    return _populate_token_with_cache(profile.current().csp, force_refresh, verbose)


def refresh_oauth_token(old_oauth_token: dict, csp_url: str):
    with OAuth2Client(token=old_oauth_token) as client:
        log.debug("Refresh auth token...")
        token_url = csp_url + "/csp/gateway/am/api/auth/token"
        from .login_support import identify_client_id

        csp_specific_req_not_oauth_standard = (identify_client_id(csp_url), "")
        new_token = client.refresh_token(token_url, auth=csp_specific_req_not_oauth_standard)
        log.debug(f"New auth token: {new_token}")
        if not new_token:
            raise Exception("CSP auth refresh failed.")
        if "cspErrorCode" in new_token:
            raise Exception(f"CSP auth failed: {new_token.get('message')}")

    return new_token


def _has_credential(auth_config: dict):
    return (
        auth_config.get("apiToken")
        or auth_config.get("api_token")
        or auth_config.get("clientId")
        or auth_config.get("client_id")
        or auth_config.get("basic")
    )


def get_auth_cache(auth_config: dict):
    cache, hash, token = _get_auth_cache(auth_config)
    return token


def _get_auth_cache(auth_config: dict):
    text = json.dumps(auth_config, default=vars)
    hash = hashlib.md5(text.encode("ascii"), usedforsecurity=False).hexdigest()
    auth_cache = profile.auth.get()
    token = auth_cache.get(hash, None)
    if token and not _is_auth_valid(token):
        token = None
    return auth_cache, hash, token


def save_auth_cache(auth_config: dict, token: dict):
    """Save the auth token to the profile auth cache."""
    cache, hash, _ = _get_auth_cache(auth_config)
    if not token:
        if hash in cache:
            del cache[hash]
        return
    if not token.get("expires_at"):
        token["expires_at"] = int(time.time() + token["expires_in"])
    cache[hash] = token
    profile.auth.set(cache)


def _populate_token_with_cache(auth_config: dict, force_refresh: bool = False, verbose: bool = False):
    if verbose:
        print("_populate_token_with_cache")
    with _login_lock:
        cache, hash, token = _get_auth_cache(auth_config)
        if token and not force_refresh:
            if verbose:
                print("Using cached auth token.")
            return token

        # invalid token. Refresh or recreate it.
        if token:
            # try using refresh token if possible
            if auth_config.get("provider", "vmwarecsp") == "vmwarecsp":
                if verbose:
                    print("Provider: vmwarecsp.")
                try:
                    token = refresh_oauth_token(token, auth_config.url)
                except Exception as e:
                    log.debug(f"Failed to refresh OAuth token: {e}")
                    token = None
            else:
                # hcs auth-service. Does not support refresh token.
                if verbose:
                    print("Provider: hcs auth-service.")
                token = None

        if not token:
            if _has_credential(auth_config):
                if verbose:
                    print("Config:", auth_config)
                token = CspClient.create(**auth_config).oauth_token()
                if verbose:
                    print("Token:", json.dumps(token, indent=4))
            else:
                if auth_config.get("browser"):
                    from .login_support import login_via_browser

                    token = login_via_browser(auth_config.url, auth_config.orgId)
                    if not token:
                        raise CtxpException("Browser auth failed.")
                else:
                    raise CtxpException("Browser auth was never attempted and no client credentials or API token provided.")

        if not token.get("expires_at"):
            token["expires_at"] = int(time.time() + token["expires_in"])

        cache[hash] = token
        profile.auth.set(cache)
    return token


class CustomOAuth2Client(OAuth2Client):
    def __init__(self, auth_config: dict):
        super().__init__()
        self.auth_config = auth_config

    def ensure_token(self):
        # pylint: disable=access-member-before-definition
        if self.token is None or not super().ensure_active_token():
            self.token = _populate_token_with_cache(self.auth_config)


def oauth_client(auth_config: dict = None):
    if not auth_config:
        auth_config = profile.current().csp
    return CustomOAuth2Client(auth_config)


def details(get_org_details: bool = False) -> dotdict:
    """Get the auth details, for the current profile"""
    oauth_token = login()
    if not oauth_token:
        return
    return details_from_token(oauth_token, get_org_details)


def details_from_token(oauth_token, get_org_details: bool = False):
    decoded = jwt.decode(oauth_token["access_token"], options={"verify_signature": False})
    org_id = decoded["context_name"]
    ret = {"token": oauth_token, "jwt": decoded, "org": {"id": org_id}}

    if get_org_details:
        csp_client = CspClient(url=profile.current().csp.url, oauth_token=oauth_token)
        try:
            org_details = csp_client.get_org_details(org_id)
        except Exception as e:
            org_details = {"error": f"Fail retrieving org details: {e}"}
        ret["org"].update(org_details)
    return dotify(ret)


def get_org_id_from_token(oauth_token: str) -> str:
    decoded = jwt.decode(oauth_token["access_token"], options={"verify_signature": False})
    return decoded["context_name"]
